# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import hashlib
import math
import os
import base64
import inspect


def create_random_name(prefix='aztest', length=24):
    if len(prefix) > length:
        raise ValueError('The length of the prefix must not be longer than random name length')

    padding_size = length - len(prefix)
    if padding_size < 4:
        raise ValueError('The randomized part of the name is shorter than 4, which may not be able to offer enough '
                         'randomness')

    random_bytes = os.urandom(int(math.ceil(float(padding_size) / 8) * 5))
    random_padding = base64.b32encode(random_bytes)[:padding_size]

    return str(prefix + random_padding.decode().lower())


def get_sha1_hash(file_path):
    sha1 = hashlib.sha256()
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(65536)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest()


def _get_content_type(entity):
    # 'headers' is a field of 'request', but it is a dict-key in 'response'
    headers = getattr(entity, 'headers', None)
    if headers is None:
        headers = entity.get('headers')

    content_type = None
    if headers:
        content_type = headers.get('content-type', None)
        if content_type:
            # content-type could an array from response, let us extract it out
            content_type = content_type[0] if isinstance(content_type, list) else content_type
            content_type = content_type.split(";")[0].lower()
    return content_type


def is_text_payload(entity):
    text_content_list = ['application/json', 'application/xml', 'text/', 'application/test-content']

    content_type = _get_content_type(entity)
    if content_type:
        return any(content_type.startswith(x) for x in text_content_list)
    return True


def is_json_payload(entity):
    return _get_content_type(entity) == 'application/json'


def trim_kwargs_from_test_function(fn, kwargs):
    # the next function is the actual test function. the kwargs need to be trimmed so
    # that parameters which are not required will not be passed to it.
    if not is_preparer_func(fn):
        args, _, kw, _ = inspect.getargspec(fn)  # pylint: disable=deprecated-method
        if kw is None:
            args = set(args)
            for key in [k for k in kwargs if k not in args]:
                del kwargs[key]


def is_preparer_func(fn):
    return getattr(fn, '__is_preparer', False)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import print_function
import unittest
import os
import inspect
import tempfile
import shutil
import logging
import threading
import six
import vcr

from .config import TestConfig
from .const import ENV_TEST_DIAGNOSE
from .utilities import create_random_name
from .decorators import live_only


class IntegrationTestBase(unittest.TestCase):
    def __init__(self, method_name):
        super(IntegrationTestBase, self).__init__(method_name)
        self.diagnose = os.environ.get(ENV_TEST_DIAGNOSE, None) == 'True'
        self.logger = logging.getLogger('azure_devtools.scenario_tests')

    def create_random_name(self, prefix, length):  # pylint: disable=no-self-use
        return create_random_name(prefix=prefix, length=length)

    def create_temp_file(self, size_kb, full_random=False):
        """
        Create a temporary file for testing. The test harness will delete the file during tearing down.
        :param float size_kb: specify the generated file size in kb.
        """
        fd, path = tempfile.mkstemp()
        os.close(fd)
        self.addCleanup(lambda: os.remove(path))

        with open(path, mode='r+b') as f:
            if full_random:
                chunk = os.urandom(1024)
            else:
                chunk = bytearray([0] * 1024)
            for _ in range(int(size_kb)):
                f.write(chunk)
            chunk = os.urandom(int(1024 * (size_kb % 1)))
            f.write(chunk)

        return path

    def create_temp_dir(self):
        """
        Create a temporary directory for testing. The test harness will delete the directory during tearing down.
        """
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir, ignore_errors=True))

        return temp_dir

    @classmethod
    def set_env(cls, key, val):
        os.environ[key] = val

    @classmethod
    def pop_env(cls, key):
        return os.environ.pop(key, None)


@live_only()
class LiveTest(IntegrationTestBase):
    pass


class ReplayableTest(IntegrationTestBase):  # pylint: disable=too-many-instance-attributes
    FILTER_HEADERS = [
        'authorization',
        'client-request-id',
        'retry-after',
        'x-ms-client-request-id',
        'x-ms-correlation-request-id',
        'x-ms-ratelimit-remaining-subscription-reads',
        'x-ms-request-id',
        'x-ms-routing-request-id',
        'x-ms-gateway-service-instanceid',
        'x-ms-ratelimit-remaining-tenant-reads',
        'x-ms-served-by',
        'x-ms-authorization-auxiliary'
    ]

    def __init__(self,  # pylint: disable=too-many-arguments
                 method_name, config_file=None, recording_dir=None, recording_name=None, recording_processors=None,
                 replay_processors=None, recording_patches=None, replay_patches=None):
        super(ReplayableTest, self).__init__(method_name)

        self.recording_processors = recording_processors or []
        self.replay_processors = replay_processors or []

        self.recording_patches = recording_patches or []
        self.replay_patches = replay_patches or []

        self.config = TestConfig(config_file=config_file)

        self.disable_recording = False

        test_file_path = inspect.getfile(self.__class__)
        recording_dir = recording_dir or os.path.join(os.path.dirname(test_file_path), 'recordings')
        self.is_live = self.config.record_mode

        self.vcr = vcr.VCR(
            cassette_library_dir=recording_dir,
            before_record_request=self._process_request_recording,
            before_record_response=self._process_response_recording,
            decode_compressed_response=True,
            record_mode='once' if not self.is_live else 'all',
            filter_headers=self.FILTER_HEADERS
        )
        self.vcr.register_matcher('query', self._custom_request_query_matcher)

        self.recording_file = os.path.join(
            recording_dir,
            '{}.yaml'.format(recording_name or method_name)
        )
        if self.is_live and os.path.exists(self.recording_file):
            os.remove(self.recording_file)

        self.in_recording = self.is_live or not os.path.exists(self.recording_file)
        self.test_resources_count = 0
        self.original_env = os.environ.copy()

    def setUp(self):
        super(ReplayableTest, self).setUp()

        # set up cassette
        cm = self.vcr.use_cassette(self.recording_file)
        self.cassette = cm.__enter__()
        self.addCleanup(cm.__exit__)

        # set up mock patches
        if self.in_recording:
            for patch in self.recording_patches:
                patch(self)
        else:
            for patch in self.replay_patches:
                patch(self)

    def tearDown(self):
        os.environ = self.original_env
        # Autorest.Python 2.x
        assert not [t for t in threading.enumerate() if t.name.startswith("AzureOperationPoller")], \
            "You need to call 'result' or 'wait' on all AzureOperationPoller you have created"
        # Autorest.Python 3.x
        assert not [t for t in threading.enumerate() if t.name.startswith("LROPoller")], \
            "You need to call 'result' or 'wait' on all LROPoller you have created"

    def _process_request_recording(self, request):
        if self.disable_recording:
            return None

        if self.in_recording:
            for processor in self.recording_processors:
                request = processor.process_request(request)
                if not request:
                    break
        else:
            for processor in self.replay_processors:
                request = processor.process_request(request)
                if not request:
                    break

        return request

    def _process_response_recording(self, response):
        from .utilities import is_text_payload
        if self.in_recording:
            # make header name lower case and filter unwanted headers
            headers = {}
            for key in response['headers']:
                if key.lower() not in self.FILTER_HEADERS:
                    headers[key.lower()] = response['headers'][key]
            response['headers'] = headers

            body = response['body']['string']
            if is_text_payload(response) and body and not isinstance(body, six.string_types):
                response['body']['string'] = body.decode('utf-8')

            for processor in self.recording_processors:
                response = processor.process_response(response)
                if not response:
                    break
        else:
            for processor in self.replay_processors:
                response = processor.process_response(response)
                if not response:
                    break

        return response

    @classmethod
    def _custom_request_query_matcher(cls, r1, r2):
        """ Ensure method, path, and query parameters match. """
        from six.moves.urllib_parse import urlparse, parse_qs  # pylint: disable=import-error, relative-import

        url1 = urlparse(r1.uri)
        url2 = urlparse(r2.uri)

        q1 = parse_qs(url1.query)
        q2 = parse_qs(url2.query)
        shared_keys = set(q1.keys()).intersection(set(q2.keys()))

        if len(shared_keys) != len(q1) or len(shared_keys) != len(q2):
            return False

        for key in shared_keys:
            if q1[key][0].lower() != q2[key][0].lower():
                return False

        return True



from pathlib import Path
import tempfile

from git import Repo

from azure_devtools.ci_tools.git_tools import (
    do_commit,
    get_files_in_commit
)

def test_do_commit():
    finished = False # Authorize PermissionError on cleanup
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            Repo.clone_from('https://github.com/lmazuel/TestingRepo.git', temp_dir)
            repo = Repo(temp_dir)

            result = do_commit(repo, 'Test {hexsha}', 'testing', 'fakehexsha')
            assert not result
            assert 'fakehexsha' not in repo.head.commit.message
            assert repo.active_branch.name == 'master'

            file_path = Path(temp_dir, 'file.txt')
            file_path.write_text('Something')

            result = do_commit(repo, 'Test {hexsha}', 'testing', 'fakehexsha')
            assert result
            assert repo.head.commit.message == 'Test fakehexsha'
            assert repo.active_branch.name == 'testing'
            assert 'file.txt' in repo.head.commit.stats.files

            file_path.write_text('New content')

            result = do_commit(repo, 'Now it is {hexsha}', 'newbranch', 'new-fakehexsha')
            assert result
            assert repo.head.commit.message == 'Now it is new-fakehexsha'
            assert repo.active_branch.name == 'newbranch'
            assert 'file.txt' in repo.head.commit.stats.files

            file_path.unlink()
            file_path.write_text('New content')

            result = do_commit(repo, 'Now it is {hexsha}', 'fakebranch', 'hexsha_not_used')
            assert not result
            assert repo.head.commit.message == 'Now it is new-fakehexsha'
            assert repo.active_branch.name == 'newbranch'

            finished = True
    except PermissionError:
        if finished:
            return
        raise

def test_get_files_in_commit():
    finished = False # Authorize PermissionError on cleanup
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            Repo.clone_from('https://github.com/lmazuel/swagger-to-sdk.git', temp_dir)
            
            files = get_files_in_commit(temp_dir, "b40451e55b26e3db61ea17bd751181cbf91f60c5")

            assert files == [
                'swaggertosdk/generate_package.py',
                'swaggertosdk/python_sdk_tools.py',
                'swaggertosdk/restapi/sdkbot.py'
            ]

            finished = True
    except PermissionError:
        if finished:
            return
        raise



import os.path
from unittest import mock

import pytest
from github.tests import Framework

from azure_devtools.ci_tools.bot_framework import BotHandler, order, build_from_issue_comment, build_from_issues

class BotFrameworkTest(Framework.TestCase):

    def setUp(self):
        self.maxDiff = None # Big diff to come
        self.recordMode = False  # turn to True to record
        self.tokenAuthMode = True
        self.replayDataFolder = os.path.join(os.path.dirname(__file__), "ReplayData")
        super(BotFrameworkTest, self).setUp()

    def test_webhook_data(self):
        github_token = self.oauth_token
        repo = self.g.get_repo("lmazuel/TestingRepo")

        fake_webhook = {
            'action': 'opened',  # What is the comment state?
            'repository': {
                'full_name': repo.full_name  # On what repo is this command?
            },
            'issue': {
                'number': 16,  # On what issue is this comment?
                'body': "@AutorestCI help"  # Message?
            },
            'sender': {
                'login': "lmazuel"  # Who wrote the command?
            },
        }
        webhook_data = build_from_issues(github_token, fake_webhook)
        assert webhook_data.text == "@AutorestCI help"
        assert webhook_data.issue.number == 16
        assert webhook_data.repo.full_name == repo.full_name

        fake_webhook = {
            'action': 'created',  # What is the comment state?
            'repository': {
                'full_name': repo.full_name  # On what repo is this command?
            },
            'issue': {
                'number': 16  # On what issue is this comment?
            },
            'sender': {
                'login': "lmazuel"  # Who wrote the command?
            },
            'comment': {
                'id': 365120206,
                'body': "@AutorestCI help"  # Message?
            }
        }
        webhook_data = build_from_issue_comment(github_token, fake_webhook)
        assert webhook_data.text == "@AutorestCI help"
        assert webhook_data.issue.number == 16
        assert webhook_data.repo.full_name == repo.full_name

    def test_bot_help(self):
        github_token = self.oauth_token
        repo = self.g.get_repo("lmazuel/TestingRepo")
        issue = repo.get_issue(16)

        class BotHelp:
            @order
            def command1(self, issue):
                pass
            @order
            def command2(self, issue):
                pass
            def notacommand(self):
                pass

        bot = BotHandler(BotHelp(), "AutorestCI", github_token)

        fake_webhook = {
            'action': 'opened',  # What is the comment state?
            'repository': {
                'full_name': issue.repository.full_name  # On what repo is this command?
            },
            'issue': {
                'number': issue.number,  # On what issue is this comment?
                'body': "@AutorestCI help"  # Message?
            },
            'sender': {
                'login': "lmazuel"  # Who wrote the command?
            },
        }

        response = bot.issues(fake_webhook)
        assert "this help message" in response["message"]
        assert "command1" in response["message"]
        assert "notacommand" not in response["message"]

        help_comment = list(issue.get_comments())[-1]
        assert "this help message" in help_comment.body
        assert "command1" in help_comment.body
        assert "notacommand" not in help_comment.body

        # Clean
        help_comment.delete()

    def test_bot_basic_command(self):
        github_token = self.oauth_token
        repo = self.g.get_repo("lmazuel/TestingRepo")
        issue = repo.get_issue(17)

        class BotCommand:
            @order
            def command1(self, issue, param1):
                assert issue.number == 17
                return "I did something with "+param1

        bot = BotHandler(BotCommand(), "AutorestCI", github_token)

        fake_webhook = {
            'action': 'opened',  # What is the comment state?
            'repository': {
                'full_name': issue.repository.full_name  # On what repo is this command?
            },
            'issue': {
                'number': issue.number,  # On what issue is this comment?
                'body': "@AutorestCI command1 myparameter"  # Message?
            },
            'sender': {
                'login': "lmazuel"  # Who wrote the command?
            },
        }

        response = bot.issues(fake_webhook)
        assert response["message"] == "I did something with myparameter"

        help_comment = list(issue.get_comments())[-1]
        assert "I did something with myparameter" in help_comment.body

        # Clean
        help_comment.delete()

    @mock.patch('traceback.format_exc')
    def test_bot_basic_failure(self, format_exc):
        format_exc.return_value = 'something to do with an exception'

        github_token = self.oauth_token
        repo = self.g.get_repo("lmazuel/TestingRepo")
        issue = repo.get_issue(18)

        class BotCommand:
            @order
            def command1(self, issue, param1):
                assert issue.number == 18
                raise ValueError("Not happy")

        bot = BotHandler(BotCommand(), "AutorestCI", github_token)

        fake_webhook = {
            'action': 'opened',  # What is the comment state?
            'repository': {
                'full_name': issue.repository.full_name  # On what repo is this command?
            },
            'issue': {
                'number': issue.number,  # On what issue is this comment?
                'body': "@AutorestCI command1 myparameter"  # Message?
            },
            'sender': {
                'login': "lmazuel"  # Who wrote the command?
            },
        }

        response = bot.issues(fake_webhook)
        assert response['message'] == 'Nothing for me or exception'

        help_comment = list(issue.get_comments())[-1]
        assert "something to do with an exception" in help_comment.body
        assert "```python" in help_comment.body

        # Clean
        help_comment.delete()


    def test_bot_unknown_command(self):
        github_token = self.oauth_token
        repo = self.g.get_repo("lmazuel/TestingRepo")
        issue = repo.get_issue(19)

        class BotCommand:
            pass

        bot = BotHandler(BotCommand(), "AutorestCI", github_token)

        fake_webhook = {
            'action': 'opened',  # What is the comment state?
            'repository': {
                'full_name': issue.repository.full_name  # On what repo is this command?
            },
            'issue': {
                'number': issue.number,  # On what issue is this comment?
                'body': "@AutorestCI command1 myparameter"  # Message?
            },
            'sender': {
                'login': "lmazuel"  # Who wrote the command?
            },
        }

        response = bot.issues(fake_webhook)
        assert "I didn't understand your command" in response['message']
        assert "command1 myparameter" in response['message']

        help_comment = list(issue.get_comments())[-1]
        assert "I didn't understand your command" in help_comment.body
        assert "command1 myparameter" in help_comment.body

        # Clean
        help_comment.delete()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# Replaced mock values
MOCKED_SUBSCRIPTION_ID = '00000000-0000-0000-0000-000000000000'
MOCKED_TENANT_ID = '00000000-0000-0000-0000-000000000000'

# Configuration environment variable
ENV_COMMAND_COVERAGE = 'AZURE_TEST_COMMAND_COVERAGE'
ENV_LIVE_TEST = 'AZURE_TEST_RUN_LIVE'
ENV_SKIP_ASSERT = 'AZURE_TEST_SKIP_ASSERT'
ENV_TEST_DIAGNOSE = 'AZURE_TEST_DIAGNOSE'



"""Github tools.
"""
from contextlib import contextmanager
import logging
import os
from pathlib import Path
import shutil
import stat
from subprocess import CalledProcessError
import traceback
from urllib.parse import urlsplit, urlunsplit

from github import Github, GithubException
from git import Repo

from .git_tools import (
    clone_to_path as _git_clone_to_path,
    checkout_with_fetch
)

_LOGGER = logging.getLogger(__name__)

class ExceptionContext:  # pylint: disable=too-few-public-methods
    def __init__(self):
        self.comment = None

@contextmanager
def exception_to_github(github_obj_to_comment, summary=""):
    """If any exception comes, log them in the given Github obj.
    """
    context = ExceptionContext()
    try:
        yield context
    except Exception:  # pylint: disable=broad-except
        if summary:
            summary = ": ({})".format(summary)
        error_type = "an unknown error"
        try:
            raise
        except CalledProcessError as err:
            error_type = "a Subprocess error"
            content = "Command: {}\n".format(err.cmd)
            content += "Finished with return code {}\n".format(err.returncode)
            if err.output:
                content += "and output:\n```shell\n{}\n```".format(err.output)
            else:
                content += "and no output"
        except Exception:  # pylint: disable=broad-except
            content = "```python\n{}\n```".format(traceback.format_exc())
        response = "<details><summary>Encountered {}{}</summary><p>\n\n".format(
            error_type,
            summary
        )
        response += content
        response += "\n\n</p></details>"
        context.comment = create_comment(github_obj_to_comment, response)

def user_from_token(gh_token):
    """Get user login from GitHub token"""
    github_con = Github(gh_token)
    return github_con.get_user()

def create_comment(github_object, body):
    """Create a comment, whatever the object is a PR, a commit or an issue.
    """
    try:
        return github_object.create_issue_comment(body)  # It's a PR
    except AttributeError:
        return github_object.create_comment(body)   # It's a commit/issue

def get_comments(github_object):
    """Get a list of comments, whater the object is a PR, a commit or an issue.
    """
    try:
        return github_object.get_issue_comments()  # It's a PR
    except AttributeError:
        return github_object.get_comments()   # It's a commit/issue

def get_files(github_object):
    """Get files from a PR or a commit.
    """
    try:
        return github_object.get_files() # Try as a PR object
    except AttributeError:
        return github_object.files # Try as a commit object

def configure_user(gh_token, repo):
    """git config --global user.email "you@example.com"
       git config --global user.name "Your Name"
    """
    user = user_from_token(gh_token)
    repo.git.config('user.email', user.email or 'adxpysdk@microsoft.com')
    repo.git.config('user.name', user.name or 'SwaggerToSDK Automation')

def get_full_sdk_id(gh_token, sdk_git_id):
    """If the SDK git id is incomplete, try to complete it with user login"""
    if not '/' in sdk_git_id:
        login = user_from_token(gh_token).login
        return '{}/{}'.format(login, sdk_git_id)
    return sdk_git_id

def sync_fork(gh_token, github_repo_id, repo, push=True):
    """Sync the current branch in this fork against the direct parent on Github"""
    if not gh_token:
        _LOGGER.warning('Skipping the upstream repo sync, no token')
        return
    _LOGGER.info('Check if repo has to be sync with upstream')
    github_con = Github(gh_token)
    github_repo = github_con.get_repo(github_repo_id)

    if not github_repo.parent:
        _LOGGER.warning('This repo has no upstream')
        return

    upstream_url = 'https://github.com/{}.git'.format(github_repo.parent.full_name)
    upstream = repo.create_remote('upstream', url=upstream_url)
    upstream.fetch()
    active_branch_name = repo.active_branch.name
    if not active_branch_name in repo.remotes.upstream.refs:
        _LOGGER.info('Upstream has no branch %s to merge from', active_branch_name)
        return
    else:
        _LOGGER.info('Merge from upstream')
    msg = repo.git.rebase('upstream/{}'.format(repo.active_branch.name))
    _LOGGER.debug(msg)
    if push:
        msg = repo.git.push()
        _LOGGER.debug(msg)

def get_or_create_pull(github_repo, title, body, head, base, *, none_if_no_commit=False):
    """Try to create the PR. If the PR exists, try to find it instead. Raises otherwise.

    You should always use the complete head syntax "org:branch", since the syntax is required
    in case of listing.

    if "none_if_no_commit" is set, return None instead of raising exception if the problem
    is that head and base are the same.
    """
    try: # Try to create or get a PR
        return github_repo.create_pull(
            title=title,
            body=body,
            head=head,
            base=base
        )
    except GithubException as err:
        err_message = err.data['errors'][0].get('message', '')
        if err.status == 422 and err_message.startswith('A pull request already exists'):
            _LOGGER.info('PR already exists, get this PR')
            return list(github_repo.get_pulls(
                head=head,
                base=base
            ))[0]
        elif none_if_no_commit and err.status == 422 and err_message.startswith('No commits between'):
            _LOGGER.info('No PR possible since head %s and base %s are the same',
                         head,
                         base)
            return None
        else:
            _LOGGER.warning("Unable to create PR:\n%s", err.data)
            raise
    except Exception as err:
        response = traceback.format_exc()
        _LOGGER.warning("Unable to create PR:\n%s", response)
        raise

def clone_to_path(gh_token, folder, sdk_git_id, branch_or_commit=None, *, pr_number=None):
    """Clone the given repo_id to the folder.

    If PR number is specified fetch the magic branches
    pull/<id>/head or pull/<id>/merge from Github. "merge" is tried first, and fallback to "head".
    Beware that pr_number implies detached head, and then no push is possible.

    If branch is specified, checkout this branch or commit finally.

    :param str branch_or_commit: If specified, switch to this branch/commit.
    :param int pr_number: PR number.
    """
    _LOGGER.info("Clone SDK repository %s", sdk_git_id)
    url_parsing = urlsplit(sdk_git_id)
    sdk_git_id = url_parsing.path
    if sdk_git_id.startswith("/"):
        sdk_git_id = sdk_git_id[1:]

    credentials_part = ''
    if gh_token:
        login = user_from_token(gh_token).login
        credentials_part = '{user}:{token}@'.format(
            user=login,
            token=gh_token
        )
    else:
        _LOGGER.warning('Will clone the repo without writing credentials')

    https_authenticated_url = 'https://{credentials}github.com/{sdk_git_id}.git'.format(
        credentials=credentials_part,
        sdk_git_id=sdk_git_id
    )
    # Clone the repo
    _git_clone_to_path(https_authenticated_url, folder)
    # If this is a PR, do some fetch to improve the number of SHA1 available
    if pr_number:
        try:
            checkout_with_fetch(folder, "pull/{}/merge".format(pr_number))
            return
        except Exception:  # pylint: disable=broad-except
            pass  # Assume "merge" doesn't exist anymore, fetch "head"
        checkout_with_fetch(folder, "pull/{}/head".format(pr_number))
    # If there is SHA1, checkout it. If PR number was given, SHA1 could be inside that PR.
    if branch_or_commit:
        repo = Repo(str(folder))
        repo.git.checkout(branch_or_commit)

def do_pr(gh_token, sdk_git_id, sdk_pr_target_repo_id, branch_name, base_branch, pr_body=""):  # pylint: disable=too-many-arguments
    "Do the PR"
    if not gh_token:
        _LOGGER.info('Skipping the PR, no token found')
        return None
    if not sdk_pr_target_repo_id:
        _LOGGER.info('Skipping the PR, no target repo id')
        return None

    github_con = Github(gh_token)
    sdk_pr_target_repo = github_con.get_repo(sdk_pr_target_repo_id)

    if '/' in sdk_git_id:
        sdk_git_owner = sdk_git_id.split('/')[0]
        _LOGGER.info("Do the PR from %s", sdk_git_owner)
        head_name = "{}:{}".format(sdk_git_owner, branch_name)
    else:
        head_name = branch_name
        sdk_git_repo = github_con.get_repo(sdk_git_id)
        sdk_git_owner = sdk_git_repo.owner.login

    try:
        github_pr = sdk_pr_target_repo.create_pull(
            title='Automatic PR from {}'.format(branch_name),
            body=pr_body,
            head=head_name,
            base=base_branch
        )
    except GithubException as err:
        if err.status == 422 and err.data['errors'][0].get('message', '').startswith('A pull request already exists'):
            matching_pulls = sdk_pr_target_repo.get_pulls(base=base_branch, head=sdk_git_owner+":"+head_name)
            matching_pull = matching_pulls[0]
            _LOGGER.info('PR already exists: %s', matching_pull.html_url)
            return matching_pull
        raise
    _LOGGER.info("Made PR %s", github_pr.html_url)
    return github_pr


def remove_readonly(func, path, _):
    "Clear the readonly bit and reattempt the removal"
    os.chmod(path, stat.S_IWRITE)
    func(path)

@contextmanager
def manage_git_folder(gh_token, temp_dir, git_id, *, pr_number=None):
    """Context manager to avoid readonly problem while cleanup the temp dir.

    If PR number is given, use magic branches "pull" from Github.
    """
    _LOGGER.debug("Git ID %s", git_id)
    if Path(git_id).exists():
        yield git_id
        return  # Do not erase a local folder, just skip here

    # Clone the specific branch
    split_git_id = git_id.split("@")
    branch = split_git_id[1] if len(split_git_id) > 1 else None
    clone_to_path(gh_token, temp_dir, split_git_id[0], branch_or_commit=branch, pr_number=pr_number)
    try:
        yield temp_dir
        # Pre-cleanup for Windows http://bugs.python.org/issue26660
    finally:
        _LOGGER.debug("Preclean Rest folder")
        shutil.rmtree(temp_dir, onerror=remove_readonly)


class GithubLink:
    def __init__(self, gitid, link_type, branch_or_commit, path, token=None):  # pylint: disable=too-many-arguments
        self.gitid = gitid
        self.link_type = link_type
        self.branch_or_commit = branch_or_commit
        self.path = path
        self.token = token

    @classmethod
    def from_string(cls, github_url):
        parsed = urlsplit(github_url)
        netloc = parsed.netloc
        if "@" in netloc:
            token, netloc = netloc.split("@")
        else:
            token = None

        split_path = parsed.path.split("/")
        split_path.pop(0)  # First is always empty
        gitid = split_path.pop(0) + "/" + split_path.pop(0)
        link_type = split_path.pop(0) if netloc != "raw.githubusercontent.com" else "raw"
        branch_or_commit = split_path.pop(0)
        path = "/".join(split_path)
        return cls(gitid, link_type, branch_or_commit, path, token)

    def __repr__(self):
        if self.link_type == "raw":
            netloc = "raw.githubusercontent.com"
            path = "/".join(["", self.gitid, self.branch_or_commit, self.path])
            # If raw and token, needs to be passed with "Authorization: token <token>", so nothing to do here
        else:
            netloc = "github.com" if not self.token else self.token + "@github.com"
            path = "/".join(["", self.gitid, self.link_type, self.branch_or_commit, self.path])
        return urlunsplit(("https", netloc, path, '', ''))

    def as_raw_link(self):
        """Returns a GithubLink to a raw content.
        """
        if self.link_type == "raw":
            return self # Can be discussed if we need an hard copy, or fail
        if self.link_type != "blob":
            raise ValueError("Cannot get a download link from a tree link")
        return self.__class__(
            self.gitid,
            "raw",
            self.branch_or_commit,
            self.path,
            self.token
        )

class DashboardCommentableObject:  # pylint: disable=too-few-public-methods
    def __init__(self, issue_or_pr, header):
        self._issue_or_pr = issue_or_pr
        self._header = header

    def create_comment(self, text):
        """Mimic issue API, so we can use it everywhere.
        Return dashboard comment.
        """
        return DashboardComment.get_or_create(self._issue_or_pr, self._header, text)

class DashboardComment:
    def __init__(self, github_comment, header):
        self.github_comment = github_comment
        self._header = header

    @classmethod
    def get_or_create(cls, issue, header, text=None):
        """Get or create the dashboard comment in this issue.
        """
        for comment in get_comments(issue):
            try:
                if comment.body.splitlines()[0] == header:
                    obj = cls(comment, header)
                    break
            except IndexError: # The comment body is empty
                pass
        # Hooooooo, no dashboard comment, let's create one
        else:
            comment = create_comment(issue, header)
            obj = cls(comment, header)
        if text:
            obj.edit(text)
        return obj

    def edit(self, text):
        self.github_comment.edit(self._header+"\n"+text)

    @property
    def body(self):
        return self.github_comment.body[len(self._header+"\n"):]

    def delete(self):
        self.github_comment.delete()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from .exceptions import AzureTestError


def patch_time_sleep_api(unit_test):
    def _time_sleep_skip(*_):
        return

    mock_in_unit_test(unit_test, 'time.sleep', _time_sleep_skip)


def patch_long_run_operation_delay(unit_test):
    def _shortcut_long_run_operation(*args, **kwargs):  # pylint: disable=unused-argument
        return

    mock_in_unit_test(unit_test,
                      'msrestazure.azure_operation.AzureOperationPoller._delay',
                      _shortcut_long_run_operation)


def mock_in_unit_test(unit_test, target, replacement):
    try:
        import unittest.mock as mock
    except ImportError:
        import mock
    import unittest

    if not isinstance(unit_test, unittest.TestCase):
        raise AzureTestError('Patches can be only called from a unit test')

    mp = mock.patch(target, replacement)
    mp.__enter__()
    unit_test.addCleanup(mp.__exit__, None, None, None)



"""Pure git tools for managing local folder Git.
"""
import logging

from git import Repo, GitCommandError

_LOGGER = logging.getLogger(__name__)

def checkout_and_create_branch(repo, name):
    """Checkout branch. Create it if necessary"""
    local_branch = repo.branches[name] if name in repo.branches else None
    if not local_branch:
        if name in repo.remotes.origin.refs:
            # If origin branch exists but not local, git.checkout is the fatest way
            # to create local branch with origin link automatically
            msg = repo.git.checkout(name)
            _LOGGER.debug(msg)
            return
        # Create local branch, will be link to origin later
        local_branch = repo.create_head(name)
    local_branch.checkout()

def checkout_create_push_branch(repo, name):
    """Checkout this branch. Create it if necessary, and push it to origin.
    """
    try:
        repo.git.checkout(name)
        _LOGGER.info("Checkout %s success", name)
    except GitCommandError:
        _LOGGER.info("Checkout %s was impossible (branch does not exist). Creating it and push it.", name)
        checkout_and_create_branch(repo, name)
        repo.git.push('origin', name, set_upstream=True)


def do_commit(repo, message_template, branch_name, hexsha):
    "Do a commit if modified/untracked files"
    repo.git.add(repo.working_tree_dir)

    if not repo.git.diff(staged=True):
        _LOGGER.warning('No modified files in this Autorest run')
        return False

    checkout_and_create_branch(repo, branch_name)
    msg = message_template.format(hexsha=hexsha)
    commit = repo.index.commit(msg)
    _LOGGER.info("Commit done: %s", msg)
    return commit.hexsha

def get_repo_hexsha(git_folder):
    """Get the SHA1 of the current repo"""
    repo = Repo(str(git_folder))
    if repo.bare:
        not_git_hexsha = "notgitrepo"
        _LOGGER.warning("Not a git repo, SHA1 used will be: %s", not_git_hexsha)
        return not_git_hexsha
    hexsha = repo.head.commit.hexsha
    _LOGGER.info("Found REST API repo SHA1: %s", hexsha)
    return hexsha

def checkout_with_fetch(git_folder, refspec, repository="origin"):
    """Fetch the refspec, and checkout FETCH_HEAD.
    Beware that you will ne in detached head mode.
    """
    _LOGGER.info("Trying to fetch and checkout %s", refspec)
    repo = Repo(str(git_folder))
    repo.git.fetch(repository, refspec)  # FETCH_HEAD should be set
    repo.git.checkout("FETCH_HEAD")
    _LOGGER.info("Fetch and checkout success for %s", refspec)

def clone_to_path(https_authenticated_url, folder, branch_or_commit=None):
    """Clone the given URL to the folder.

    :param str branch_or_commit: If specified, switch to this branch. Branch must exist.
    """
    _LOGGER.info("Cloning repo")
    repo = Repo.clone_from(https_authenticated_url, str(folder))
    # Do NOT clone and set branch at the same time, since we allow branch to be a SHA1
    # And you can't clone a SHA1
    if branch_or_commit:
        _LOGGER.info("Checkout branch_or_commit %s", branch_or_commit)
        repo.git.checkout(branch_or_commit)

    _LOGGER.info("Clone success")

def get_files_in_commit(git_folder, commit_id="HEAD"):
    """List of files in HEAD commit.
    """
    repo = Repo(str(git_folder))
    output = repo.git.diff("--name-only", commit_id+"^", commit_id)
    return output.splitlines()



import os.path
from pathlib import Path
from subprocess import CalledProcessError
import tempfile
from unittest import mock

import pytest

from git import Repo, GitCommandError
from github import GithubException
from github.tests import Framework

from azure_devtools.ci_tools.github_tools import (
    exception_to_github,
    user_from_token,
    configure_user,
    clone_to_path,
    manage_git_folder,
    do_pr,
    get_files,
    create_comment,
    GithubLink,
    DashboardCommentableObject,
    DashboardComment,
    get_or_create_pull
)

class GithubTools(Framework.TestCase):

    def setUp(self):
        self.maxDiff = None # Big diff to come
        self.recordMode = False  # turn to True to record
        self.tokenAuthMode = True
        self.replayDataFolder = os.path.join(os.path.dirname(__file__), "ReplayData")
        super(GithubTools, self).setUp()

    @mock.patch('traceback.format_exc')
    def test_exception_to_github(self, format_exc):
        format_exc.return_value = 'something to do with an exception'

        # Prepare
        repo = self.g.get_repo("lmazuel/TestingRepo")
        issue = repo.get_issue(13)

        # Act
        with exception_to_github(issue) as error:
            pass

        assert error.comment is None

        # Act
        with exception_to_github(issue) as error:
            "Test".fakemethod(12)  # pylint: disable=no-member

        # Test
        assert error.comment is not None
        assert "Encountered an unknown error" in error.comment.body

        # Clean my mess
        error.comment.delete()

        # Act
        with exception_to_github(issue, "Python bot") as error:
            "Test".fakemethod(12)  # pylint: disable=no-member

        # Test
        assert error.comment is not None
        assert "Encountered an unknown error: (Python bot)" in error.comment.body

        # Clean my mess
        error.comment.delete()

        # Act
        with exception_to_github(issue, "Python bot") as error:
            raise CalledProcessError(
                2,
                ["autorest", "readme.md"],
                "Error line 1\nError line 2"
            )

        # Test
        assert error.comment is not None
        assert "Encountered a Subprocess error: (Python bot)" in error.comment.body
        assert "Error line 1" in error.comment.body

        # Clean my mess
        error.comment.delete()

        # Act
        with exception_to_github(issue, "Python bot") as error:
            raise CalledProcessError(
                2,
                ["autorest", "readme.md"],
            )

        # Test
        assert error.comment is not None
        assert "Encountered a Subprocess error: (Python bot)" in error.comment.body
        assert "no output" in error.comment.body

        # Clean my mess
        error.comment.delete()

    def test_get_or_create_pull(self):
        repo = self.g.get_repo("lmazuel/TestingRepo")

        # b1 and b2 should exist and be the same
        with pytest.raises(GithubException):
            get_or_create_pull(repo, "Title", "Body", "b1", "b2")

        prresult = get_or_create_pull(repo, "Title", "Body", "b1", "b2", none_if_no_commit=True)
        assert prresult is None

    @mock.patch('traceback.format_exc')
    def test_dashboard(self, format_exc):
        format_exc.return_value = 'something to do with an exception'

        # Prepare
        repo = self.g.get_repo("lmazuel/TestingRepo")
        issue = repo.get_issue(15)
        initial_size = len(list(issue.get_comments()))
        header = "# MYHEADER"

        dashboard = DashboardCommentableObject(issue, header)

        with exception_to_github(dashboard, "Python bot") as error:
            "Test".fakemethod(12)  # pylint: disable=no-member

        after_size = len(list(issue.get_comments()))
        assert after_size == initial_size + 1

        assert error.comment is not None
        assert "Encountered an unknown error" in error.comment.body

        dashboard.create_comment("New text comment")
        after_size_2 = len(list(issue.get_comments()))
        assert after_size == after_size_2

        # Clean my mess
        error.comment.delete()

    def test_get_user(self):
        github_token = self.oauth_token
        user = user_from_token(github_token)
        assert user.login == 'lmazuel'

    def test_get_files(self):
        repo = self.g.get_repo("Azure/azure-sdk-for-python")
        pr = repo.get_pull(1833)
        files = get_files(pr)
        assert "azure-mgmt-consumption/azure/mgmt/consumption/consumption_management_client.py" in [f.filename for f in files]

        commit = repo.get_commit("042b7a5840ff471776bb64e46b50950ee9f84430")
        files = get_files(commit)
        assert "azure-mgmt-consumption/azure/mgmt/consumption/consumption_management_client.py" in [f.filename for f in files]

    def test_create_comment(self):
        repo = self.g.get_repo("lmazuel/TestingRepo")
        issue = repo.get_issue(14)
        comment = create_comment(issue, "This is a test")
        comment.delete()

        pull = repo.get_pull(2)
        comment = create_comment(pull, "This is a test")
        comment.delete()

    def test_configure(self):
        github_token = self.oauth_token
        finished = False
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                try:
                    Repo.clone_from('https://github.com/lmazuel/TestingRepo.git', temp_dir)
                    repo = Repo(temp_dir)

                    # If it's not throwing, I'm happy enough
                    configure_user(github_token, repo)

                    assert repo.git.config('--get', 'user.name') == 'Laurent Mazuel'
                except Exception as err:
                    print(err)
                    pytest.fail(err)
                else:
                    finished = True
        except PermissionError:
            if finished:
                return
            raise

    def test_clone_path(self):
        github_token = self.oauth_token
        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                clone_to_path(github_token, temp_dir, "lmazuel/TestingRepo")
                assert (Path(temp_dir) / Path("README.md")).exists()

                finished = True
        except PermissionError:
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                clone_to_path(github_token, temp_dir, "https://github.com/lmazuel/TestingRepo")
                assert (Path(temp_dir) / Path("README.md")).exists()

                finished = True
        except PermissionError:
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                clone_to_path(github_token, temp_dir, "lmazuel/TestingRepo", "lmazuel-patch-1")
                assert (Path(temp_dir) / Path("README.md")).exists()

                finished = True
        except PermissionError:
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                with pytest.raises(GitCommandError):
                    clone_to_path(github_token, temp_dir, "lmazuel/TestingRepo", "fakebranch")

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        # PR 2 must be open, or the test means nothing
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                clone_to_path(github_token, temp_dir, "lmazuel/TestingRepo", pr_number=2)
                assert (Path(temp_dir) / Path("README.md")).exists()

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        # PR 1 must be MERGED, or the test means nothing
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                clone_to_path(github_token, temp_dir, "lmazuel/TestingRepo", pr_number=1)
                assert (Path(temp_dir) / Path("README.md")).exists()

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        # PR 2 must be opened, or the test means nothing
        repo = self.g.get_repo("lmazuel/TestingRepo")
        pr = repo.get_pull(2)
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                clone_to_path(github_token, temp_dir, "lmazuel/TestingRepo", branch_or_commit=pr.merge_commit_sha, pr_number=2)
                assert (Path(temp_dir) / Path("README.md")).stat().st_size >= 107 # File in the PR

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                with pytest.raises(GitCommandError):
                    clone_to_path(github_token, temp_dir, "lmazuel/TestingRepo", pr_number=123456789)

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

    def test_manage_git_folder(self):
        github_token = self.oauth_token
        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir, \
                        manage_git_folder(github_token, temp_dir, "lmazuel/TestingRepo") as rest_repo:

                assert (Path(rest_repo) / Path("README.md")).exists()

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir, \
                        manage_git_folder(github_token, temp_dir, "lmazuel/TestingRepo@lmazuel-patch-1") as rest_repo:

                assert (Path(rest_repo) / Path("README.md")).exists()
                assert "lmazuel-patch-1" in str(Repo(rest_repo).active_branch)

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

        finished = False # Authorize PermissionError on cleanup
        try:
            with tempfile.TemporaryDirectory() as temp_dir, \
                        manage_git_folder(github_token, temp_dir, "lmazuel/TestingRepo", pr_number=1) as rest_repo:

                assert (Path(rest_repo) / Path("README.md")).exists()
                with pytest.raises(TypeError) as err:
                    Repo(rest_repo).active_branch
                assert "HEAD is a detached symbolic reference" in str(err)

                finished = True
        except (PermissionError, FileNotFoundError):
            if not finished:
                raise

    def test_do_pr(self):
        github_token = self.oauth_token
        # Should do nothing
        do_pr(None, 'bad', 'bad', 'bad', 'bad')

        # Should do nothing
        do_pr(github_token, 'bad', None, 'bad', 'bad')

        # FIXME - more tests


def test_github_link():
    inputstr = "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/specification/billing/resource-manager/readme.md"
    link = GithubLink.from_string(inputstr)
    assert link.gitid == "Azure/azure-rest-api-specs"
    assert link.branch_or_commit == "master"
    assert link.link_type == "raw"
    assert link.path == "specification/billing/resource-manager/readme.md"
    assert str(link) == inputstr
    raw_link = link.as_raw_link()
    assert isinstance(raw_link, GithubLink)
    assert str(raw_link) == str(link)

    inputstr = "https://github.com/Azure/azure-rest-api-specs/blob/master/specification/billing/resource-manager/readme.md"
    link = GithubLink.from_string(inputstr)
    assert link.gitid == "Azure/azure-rest-api-specs"
    assert link.branch_or_commit == "master"
    assert link.link_type == "blob"
    assert link.path == "specification/billing/resource-manager/readme.md"
    assert str(link) == inputstr
    raw_link = link.as_raw_link()
    assert isinstance(raw_link, GithubLink)
    assert str(raw_link) == "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/specification/billing/resource-manager/readme.md"

    inputstr = "https://github.com/Azure/azure-rest-api-specs/tree/master/specification/billing/resource-manager"
    link = GithubLink.from_string(inputstr)
    assert link.gitid == "Azure/azure-rest-api-specs"
    assert link.branch_or_commit == "master"
    assert link.link_type == "tree"
    assert link.path == "specification/billing/resource-manager"
    assert str(link) == inputstr
    with pytest.raises(ValueError):
        link.as_raw_link()

    inputstr = "https://token@github.com/Azure/azure-rest-api-specs/blob/master/specification/billing/resource-manager/readme.md"
    link = GithubLink.from_string(inputstr)
    assert link.token == "token"
    assert link.gitid == "Azure/azure-rest-api-specs"
    assert link.branch_or_commit == "master"
    assert link.link_type == "blob"
    assert link.path == "specification/billing/resource-manager/readme.md"
    assert str(link) == inputstr
    raw_link = link.as_raw_link()
    assert isinstance(raw_link, GithubLink)
    # Raw link with token does not use token in URL, since it has to be provided as Authorization: token <token>
    assert str(raw_link) == "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/specification/billing/resource-manager/readme.md"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
import mock
from azure_devtools.scenario_tests.utilities import (create_random_name, get_sha1_hash,
                                                     is_text_payload, is_json_payload)


class TestUtilityFunctions(unittest.TestCase):
    def test_create_random_name_default_value(self):
        default_generated_name = create_random_name()
        self.assertTrue(default_generated_name.startswith('aztest'))
        self.assertEqual(24, len(default_generated_name))
        self.assertTrue(isinstance(default_generated_name, str))

    def test_create_random_name_randomness(self):
        self.assertEqual(100, len({create_random_name() for _ in range(100)}))

    def test_create_random_name_customization(self):
        customized_name = create_random_name(prefix='pauline', length=61)
        self.assertTrue(customized_name.startswith('pauline'))
        self.assertEqual(61, len(customized_name))
        self.assertTrue(isinstance(customized_name, str))

    def test_create_random_name_exception_long_prefix(self):
        prefix = 'prefix-too-long'

        with self.assertRaises(ValueError) as cm:
            create_random_name(prefix, length=len(prefix)-1)
        self.assertEqual(str(cm.exception), 'The length of the prefix must not be longer than random name length')

        self.assertTrue(create_random_name(prefix, length=len(prefix)+4).startswith(prefix))

    def test_create_random_name_exception_not_enough_space_for_randomness(self):
        prefix = 'prefix-too-long'

        for i in range(4):
            with self.assertRaises(ValueError) as cm:
                create_random_name(prefix, length=len(prefix) + i)
            self.assertEqual(str(cm.exception), 'The randomized part of the name is shorter than 4, which may not be '
                                                'able to offer enough randomness')

    def test_get_sha1_hash(self):
        import tempfile
        with tempfile.NamedTemporaryFile() as f:
            content = b"""
All the world's a stage,
And all the men and women merely players;
They have their exits and their entrances,
And one man in his time plays many parts,
His acts being seven ages. At first, the infant,
Mewling and puking in the nurse's arms.
Then the whining schoolboy, with his satchel
And shining morning face, creeping like snail
Unwillingly to school. And then the lover,
Sighing like furnace, with a woeful ballad
Made to his mistress' eyebrow. Then a soldier,
Full of strange oaths and bearded like the pard,
Jealous in honor, sudden and quick in quarrel,
Seeking the bubble reputation
Even in the cannon's mouth. And then the justice,
In fair round belly with good capon lined,
With eyes severe and beard of formal cut,
Full of wise saws and modern instances;
And so he plays his part. The sixth age shifts
Into the lean and slippered pantaloon,
With spectacles on nose and pouch on side;
His youthful hose, well saved, a world too wide
For his shrunk shank, and his big manly voice,
Turning again toward childish treble, pipes
And whistles in his sound. Last scene of all,
That ends this strange eventful history,
Is second childishness and mere oblivion,
Sans teeth, sans eyes, sans taste, sans everything.

William Shakespeare
            """
            f.write(content)
            f.seek(0)
            hash_value = get_sha1_hash(f.name)
            self.assertEqual('1a9ea462ce80aac3f1cacbdf59d3a630df01b933593a2c53bccc25ecc2569e31', hash_value)

    def test_text_payload(self):
        http_entity = mock.MagicMock()
        headers = {}
        http_entity.headers = headers

        headers['content-type'] = 'foo/'
        self.assertFalse(is_text_payload(http_entity))

        headers['content-type'] = 'text/html; charset=utf-8'
        self.assertTrue(is_text_payload(http_entity))

        headers['content-type'] = 'APPLICATION/JSON; charset=utf-8'
        self.assertTrue(is_text_payload(http_entity))

        headers['content-type'] = 'APPLICATION/xml'
        self.assertTrue(is_text_payload(http_entity))

        http_entity.headers = None  # default to text mode if there is no header
        self.assertTrue(is_text_payload(http_entity))

    def test_json_payload(self):
        http_entity = mock.MagicMock()
        headers = {}
        http_entity.headers = headers

        headers['content-type'] = 'APPLICATION/JSON; charset=utf-8'
        self.assertTrue(is_json_payload(http_entity))

        headers['content-type'] = 'application/json; charset=utf-8'
        self.assertTrue(is_json_payload(http_entity))

        headers['content-type'] = 'application/xml; charset=utf-8'
        self.assertFalse(is_json_payload(http_entity))



"""Configuration of fixtures for pytest"""
import logging
import os
import sys
import types

from github import Github

import pytest

_LOGGER = logging.getLogger(__name__)

collect_ignore = []
if sys.version_info < (3, 6):
    # Might do something more generic later
    collect_ignore.append("test_bot_framework.py")
    collect_ignore.append("test_git_tools.py")
    collect_ignore.append("test_github_tools.py")

_context = { 
    'login': "login",
    'password': "password",
    'oauth_token': os.environ.get('GH_TOKEN', 'oauth_token')
}
_test_context_module = types.ModuleType(
    'GithubCredentials',
    'Module created to provide a context for tests'
)
_test_context_module.__dict__.update(_context)
sys.modules['GithubCredentials'] = _test_context_module


@pytest.fixture
def github_token():
    """Return the Github token to use for real tests."""
    if not 'GH_TOKEN' in os.environ:
        _LOGGER.warning('GH_TOKEN must be defined for this test')
        return "faketoken"
    return os.environ['GH_TOKEN']

@pytest.fixture
def github_client(github_token):
    """Return a Github client with configured token."""
    return Github(github_token)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import io
from setuptools import setup


VERSION = "1.2.0"


CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'License :: OSI Approved :: MIT License',
]


DEPENDENCIES = [
    'ConfigArgParse>=0.12.0',
    'six>=1.10.0',
    'vcrpy>=1.11.0',
]

with io.open('README.rst', 'r', encoding='utf-8') as f:
    README = f.read()

setup(
    name='azure-devtools',
    version=VERSION,
    description='Microsoft Azure Development Tools for SDK',
    long_description=README,
    license='MIT',
    author='Microsoft Corporation',
    author_email='ptvshelp@microsoft.com',
    url='https://github.com/Azure/azure-python-devtools',
    zip_safe=False,
    classifiers=CLASSIFIERS,
    packages=[
        'azure_devtools',
        'azure_devtools.scenario_tests',
        'azure_devtools.ci_tools',
    ],
    extras_require={
        'ci_tools':[
            "PyGithub>=1.40", # Can Merge PR after 1.36, "requests" and tests after 1.40
            "GitPython",
            "requests>=2.0"
        ]
    },
    package_dir={'': 'src'},
    install_requires=DEPENDENCIES,
)



from collections import namedtuple
from functools import lru_cache
import logging
import os
import re

from github import Github, GithubException, UnknownObjectException

from .github_tools import (
    exception_to_github,
)

_LOGGER = logging.getLogger(__name__)


def order(function):
    function.bot_order = True
    return function

WebhookMetadata = namedtuple(
    'WebhookMetadata',
    ['repo', 'issue', 'text', 'comment']
)

def build_from_issue_comment(gh_token, body):
    """Create a WebhookMetadata from a comment added to an issue.
    """
    if body["action"] in ["created", "edited"]:
        github_con = Github(gh_token)
        repo = github_con.get_repo(body['repository']['full_name'])
        issue = repo.get_issue(body['issue']['number'])
        text = body['comment']['body']
        try:
            comment = issue.get_comment(body['comment']['id'])
        except UnknownObjectException:
            # If the comment has already disapeared, skip the command
            return None
        return WebhookMetadata(repo, issue, text, comment)
    return None

def build_from_issues(gh_token, body):
    """Create a WebhookMetadata from an opening issue text.
    """
    if body["action"] in ["opened", "edited"]:
        github_con = Github(gh_token)
        repo = github_con.get_repo(body['repository']['full_name'])
        issue = repo.get_issue(body['issue']['number'])
        text = body['issue']['body']
        comment = issue  # It's where we update the comment: in the issue itself
        return WebhookMetadata(repo, issue, text, comment)
    return None

@lru_cache()
def robot_name_from_env_variable():
    github_con = Github(os.environ["GH_TOKEN"])
    return github_con.get_user().login


class BotHandler:
    def __init__(self, handler, robot_name=None, gh_token=None):
        self.handler = handler
        self.gh_token = gh_token or os.environ["GH_TOKEN"]
        self.robot_name = robot_name or robot_name_from_env_variable()

    def _is_myself(self, body):
        return body['sender']['login'].lower() == self.robot_name.lower()

    def issue_comment(self, body):
        if self._is_myself(body):
            return {'message': 'I don\'t talk to myself, I\'m not schizo'}
        webhook_data = build_from_issue_comment(self.gh_token, body)
        return self.manage_comment(webhook_data)

    def issues(self, body):
        if self._is_myself(body):
            return {'message': 'I don\'t talk to myself, I\'m not schizo'}
        webhook_data = build_from_issues(self.gh_token, body)
        return self.manage_comment(webhook_data)

    def orders(self):
        """Return method tagged "order" in the handler.
        """
        return [order_cmd for order_cmd in dir(self.handler)
                if getattr(getattr(self.handler, order_cmd), "bot_order", False)]

    def manage_comment(self, webhook_data):
        if webhook_data is None:
            return {'message': 'Nothing for me'}
        # Is someone talking to me:
        message = re.search("@{} (.*)".format(self.robot_name), webhook_data.text, re.I)
        response = None
        if message:
            command = message.group(1)
            split_text = command.lower().split()
            orderstr = split_text.pop(0)
            if orderstr == "help":
                response = self.help_order()
            elif orderstr in self.orders():
                try:  # Reaction is fun, but it's preview not prod.
                      # Be careful, don't fail the command if we can't thumbs up...
                    webhook_data.comment.create_reaction("+1")
                except GithubException:
                    pass
                with exception_to_github(webhook_data.issue):  # Just in case
                    response = getattr(self.handler, orderstr)(webhook_data.issue, *split_text)
            else:
                response = "I didn't understand your command:\n```bash\n{}\n```\nin this context, sorry :(\n".format(
                    command
                )
                response += self.help_order()
            if response:
                webhook_data.issue.create_comment(response)
                return {'message': response}
        return {'message': 'Nothing for me or exception'}

    def help_order(self):
        orders = ["This is what I can do:"]
        for orderstr in self.orders():
            orders.append("- `{}`".format(orderstr))
        orders.append("- `help` : this help message")
        return "\n".join(orders)
